require 'json'
require 'aws-sdk-s3'

require_relative '../../helpers/shell_interpolation'

module Covalence
  module S3

    ##### AWS ennvironment variables #####
    #
    # AWS_ACCESS_KEY_ID – AWS access key.
    # AWS_SECRET_ACCESS_KEY – AWS secret key. Access and secret key variables override
    #                         credentials stored in credential and config files.
    # AWS_REGION – AWS region. This variable overrides the default region of the in-use
    #              profile, if set.

    REGION = ENV['AWS_REGION']

    class Client
      def initialize(region: REGION)
        @s3 = Aws::S3::Client.new(region: region)
        self.reset_cache
      end

      def reset_cache
        @cache = Hash.new{|h,k| h[k] = Hash.new}
      end

      def get_cache
        @cache.to_s
      end

      def get_doc(bucket, document)
        @cache[bucket][document] ||= begin
          @s3.get_object(bucket: bucket, key: document).body.read
        rescue Aws::S3::Errors::ServiceError => err
          raise err, "Unable to retrieve document '#{document}' from bucket '#{bucket}'"
        end
      end

      def get_key(bucket, document, name)
        begin
          doc = self.get_doc(bucket, document)
        rescue Aws::S3::Errors::ServiceError => err
          fail err.message + " and the key is: #{name}. Check your data files."
        end

        # Parse JSON response
        begin
          parsed = JSON.parse(doc)
        rescue JSON::ParserError => err
          fail "No results or unable to parse document '#{document}': " + err.message
        end

        # Determine whether the document is a Terraform state file
        tf_state = true if parsed.has_key?('terraform_version')

        # Return ID for the key specified
        if tf_state
          tf_vers = Gem::Version.new(parsed.fetch('terraform_version'))

          if tf_vers >= Gem::Version.new('0.12.0')
            root = parsed
          else
            root = parsed.fetch('modules')[0]
          end

          outputs = root.fetch('outputs')
          return outputs.fetch(name)
        end
        return parsed.fetch(name) if parsed.has_key?(name)
        fail "Requested key '#{name}' not found"
      end
    end

    # Return configuration for remote state store.
    def self.get_state_store(params, workspace_enabled=false)
      raise "State store parameters must be a Hash" unless params.is_a?(Hash)
      required_params = [
        'bucket',
        'name'
      ]
      required_params.each do |param|
        raise "Missing '#{param}' store parameter" unless params.has_key?(param)
      end

      config = <<-CONF
terraform {
  backend "s3" {
CONF

      if !params.has_key?('key')
        Covalence::LOGGER.debug "'key' parameter not specified. Inferring value from 'name' parameter."

        if workspace_enabled
          config += "    key = \"terraform.tfstate\"\n"
        else
          config += "    key = \"#{params['name']}/terraform.tfstate\"\n"
        end
      end

      if !params.has_key?('workspace_key_prefix')
        if workspace_enabled
          Covalence::LOGGER.debug "'workspace_key_prefix' parameter not specified. Inferring value from 'name' parameter."
          config += "    workspace_key_prefix = \"#{params['name']}\"\n"
        end
      end

      params.each do |k,v|
        next if k == 'name'

        v = Covalence::Helpers::ShellInterpolation.parse_shell(v) if v.to_s.include?("$(")
        config += "    #{k} = \"#{v}\"\n"
      end

      config += "  }\n}\n"

      return config
    end

    # Return module capabilities
    # TODO: maybe a state_store mixin later
    #def self.has_key_read?
      #return true
    #end

    #def self.has_key_write?
      #return false
    #end

    #def self.has_state_read?
      #return true
    #end

    def self.has_state_store?
      return true
    end

    # Key lookups
    def self.lookup(type, params)
      raise "Lookup parameters must be a Hash" unless params.is_a?(Hash)

      case
      when type == 'key' || type == 'state'
        required_params = [
          'bucket',
          'document',
          'key'
        ]
        required_params.each do |param|
          raise "Missing '#{param}' lookup parameter" unless params.has_key?(param)
        end
        raise "Missing 'key' lookup parameter" unless params.has_key? 'key'
        client = S3::Client.new(region: REGION)
        client.get_key(params['bucket'],params['document'],params['key'])
      end
    end
  end
end
